@@ -6,7 +6,8 @@ import asyncio
 import datetime
 import json
 import logging
-from async_property import async_property, async_cached_property
+import socket
+from async_property import async_cached_property
 from types import SimpleNamespace
 
 from neohubapi.enums import ScheduleFormat
@@ -14,43 +15,74 @@ from neohubapi.enums import schedule_format_int_to_enum
 from neohubapi.neostat import NeoStat
 
 
+class Error(Exception):
+    pass
+
+
+class NeoHubUsageError(Error):
+    pass
+
+
+class NeoHubConnectionError(Error):
+    pass
+
+
 class NeoHub:
-    def __init__(self, host='Neo-Hub', port=4242):
+    def __init__(self, host='Neo-Hub', port=4242, request_timeout=5, request_attempts=1):
         self._logger = logging.getLogger('neohub')
         self._host = host
         self._port = port
+        self._request_timeout = request_timeout
+        self._request_attempts = request_attempts
 
-    async def _send(self, message, expected_reply=None):
-        reader, writer = await asyncio.open_connection(self._host, self._port)
+    async def _send_message(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, message: str):
         encoded_message = bytearray(json.dumps(message) + "\0\r", "utf-8")
         self._logger.debug(f"Sending message: {encoded_message}")
         writer.write(encoded_message)
         await writer.drain()
 
         data = await reader.readuntil(b'\0')
-        data = data.strip(b'\0')
-        json_string = data.decode('utf-8')
-        self._logger.debug(f"Received message: {json_string}")
 
         writer.close()
         await writer.wait_closed()
 
-        try:
-            reply = json.loads(json_string, object_hook=lambda d: SimpleNamespace(**d))
-        except json.decoder.JSONDecodeError as e:
-            if expected_reply is None:
-                raise(e)
-            else:
-                return False
+        data = data.strip(b'\0')
+        return data
 
-        if expected_reply is None:
-            return reply
-        else:
-            if reply.__dict__ == expected_reply:
-                return True
-            else:
-                self._logger.error(f"Unexpected reply: {reply}")
-                return False
+    async def _send(self, message, expected_reply=None):
+        last_exception = None
+        for attempt in range(1, self._request_attempts+1):
+            try:
+                reader, writer = await asyncio.open_connection(self._host, self._port)
+                data = await asyncio.wait_for(
+                        self._send_message(reader, writer, message), timeout=self._request_timeout)
+                json_string = data.decode('utf-8')
+                self._logger.debug(f"Received message: {json_string}")
+                reply = json.loads(json_string, object_hook=lambda d: SimpleNamespace(**d))
+
+                if expected_reply is None:
+                    return reply
+                if reply.__dict__ == expected_reply:
+                    return True
+                self._logger.error(f"[{attempt}] Unexpected reply: {reply}")
+            except (socket.gaierror, ConnectionRefusedError) as e:
+                last_exception = NeoHubConnectionError(e)
+                self._logger.error(f"[{attempt}] Could not connect to NeoHub at {self._host}: {e}")
+            except asyncio.TimeoutError as e:
+                last_exception = e
+                self._logger.error(f"[{attempt}] Timed out while sending a message to {self._host}")
+                if writer is not None:
+                    writer.close()
+            except json.decoder.JSONDecodeError as e:
+                last_exception = e
+                self._logger.error(f"[{attempt}] Could not decode JSON: {e}")
+            # Wait for 1/2 of the timeout value before retrying.
+            if self._request_attempts > 1 and attempt < self._request_attempts:
+                await asyncio.sleep(self._request_timeout / 2)
+
+        if expected_reply is None and last_exception is not None:
+            raise(last_exception)
+        return False
 
     async def firmware(self):
         """
@@ -111,7 +143,11 @@ class NeoHub:
         Only channels 11, 14, 15, 19, 20, 24, 25 are allowed.
         """
 
-        message = {"SET_CHANNEL": channel}
+        try:
+            message = {"SET_CHANNEL": int(channel)}
+        except ValueError:
+            raise NeoHubUsageError('channel must be a number')
+
         reply = {"result": "Trying to change channel"}
 
         result = await self._send(message, reply)
@@ -128,14 +164,16 @@ class NeoHub:
         result = await self._send(message, reply)
         return result
 
-    async def set_format(self, format: ScheduleFormat):
+    async def set_format(self, sched_format: ScheduleFormat):
         """
         Sets schedule format
 
         Format is specified using ScheduleFormat enum:
         """
+        if not isinstance(sched_format, ScheduleFormat):
+            raise NeoHubUsageError('sched_format must be a ScheduleFormat')
 
-        message = {"SET_FORMAT": format}
+        message = {"SET_FORMAT": sched_format.value}
         reply = {"result": "Format was set"}
 
         result = await self._send(message, reply)
@@ -164,6 +202,9 @@ class NeoHub:
         start: beginning of holiday
         end: end of holiday
         """
+        for datetime_arg in (start, end):
+            if not isinstance(datetime_arg, datetime.datetime):
+                raise NeoHubUsageError('start and end must be datetime.datetime objects')
 
         message = {"HOLIDAY": [start.strftime("%H%M%S%d%m%Y"), end.strftime("%H%M%S%d%m%Y")]}
 
@@ -179,8 +220,10 @@ class NeoHub:
         message = {"GET_HOLIDAY": 0}
 
         result = await self._send(message)
-        result.start = datetime.datetime.strptime(result.start.strip(), "%a %b %d %H:%M:%S %Y") if result.start else None
-        result.end = datetime.datetime.strptime(result.end.strip(), "%a %b %d %H:%M:%S %Y")  if result.end else None
+        result.start = datetime.datetime.strptime(
+                result.start.strip(), "%a %b %d %H:%M:%S %Y") if result.start else None
+        result.end = datetime.datetime.strptime(
+                result.end.strip(), "%a %b %d %H:%M:%S %Y") if result.end else None
         return result
 
     async def cancel_holiday(self):
@@ -242,39 +285,32 @@ class NeoHub:
         result = await self._send(message, reply)
         return result
 
-    async def set_date(self, date=None):
+    async def set_date(self, date: datetime.datetime = datetime.datetime.today()):
         """
         Sets current date
 
         By default, set to current date. Can be optionally passed datetime.datetime object
         """
 
-        if date is None:
-            date = datetime.datetime.today()
-
         message = {"SET_DATE": [date.year, date.month, date.day]}
         reply = {"result": "Date is set"}
 
         result = await self._send(message, reply)
         return result
 
-    async def set_time(self, time=None):
+    async def set_time(self, time: datetime.datetime = datetime.datetime.now()):
         """
         Sets current time
 
         By default, set to current time. Can be optionally passed datetime.datetime object
         """
-
-        if time is None:
-            time = datetime.datetime.now()
-
         message = {"SET_TIME": [time.hour, time.minute]}
         reply = {"result": "time set"}
 
         result = await self._send(message, reply)
         return result
 
-    async def set_datetime(self, date_time=None):
+    async def set_datetime(self, date_time: datetime.datetime = datetime.datetime.now()):
         """
         Convenience method to set both date and time
         """
@@ -295,7 +331,7 @@ class NeoHub:
         result = await self._send(message, reply)
         return result
 
-    async def set_dst(self, state: bool, region=None):
+    async def set_dst(self, state: bool, region: str = None):
         """
         Enables/disables automatic DST handling.
 
@@ -308,7 +344,7 @@ class NeoHub:
 
         valid_timezones = ["UK", "EU", "NZ"]
         if region not in valid_timezones:
-            return False
+            raise NeoHubUsageError(f'region must be in {valid_timezones}')
 
         result = await self._send(message, reply)
         return result
@@ -352,7 +388,7 @@ class NeoHub:
 
         return hub_data, devices
 
-    async def permit_join(self, name, timeout_s=120):
+    async def permit_join(self, name, timeout_s: int = 120):
         """
         Permit new thermostat to join network
 
@@ -366,7 +402,7 @@ class NeoHub:
         message = {"PERMIT_JOIN": [timeout_s, name]}
         reply = {"result": "network allows joining"}
 
-        result = await self._send(message)
+        result = await self._send(message, reply)
         return result
 
     async def set_lock(self, pin: int, devices: [NeoStat]):
@@ -376,8 +412,11 @@ class NeoHub:
         PIN is a four digit number
         """
 
-        if pin < 0 or pin > 9999:
-            return False
+        try:
+            if pin < 0 or pin > 9999:
+                return False
+        except TypeError:
+            raise NeoHubUsageError('pin must be a number')
 
         pins = []
         for x in range(4):
@@ -385,7 +424,11 @@ class NeoHub:
             pin = pin // 10
         pins.reverse()
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
+
         message = {"LOCK": [pins, names]}
         reply = {"result": "locked"}
 
@@ -397,7 +440,11 @@ class NeoHub:
         Unlocks PIN locked thermostats
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
+
         message = {"UNLOCK": names}
         reply = {"result": "unlocked"}
 
@@ -409,7 +456,11 @@ class NeoHub:
         Enables or disables Frost mode
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
+
         message = {"FROST_ON" if state else "FROST_OFF": names}
         reply = {"result": "frost on" if state else "frost off"}
 
@@ -423,7 +474,11 @@ class NeoHub:
         The temperature will be reset once next comfort level is reached
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
+
         message = {"SET_TEMP": [temperature, names]}
         reply = {"result": "temperature was set"}
 
@@ -441,7 +496,11 @@ class NeoHub:
         3: 3 degrees
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
+
         message = {"SET_DIFF": [switching_differential, names]}
         reply = {"result": "switching differential was set"}
 
@@ -453,7 +512,11 @@ class NeoHub:
         Returns time in minutes required to change temperature by 1 degree
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
+
         message = {"VIEW_ROC": names}
 
         result = await self._send(message)
@@ -467,7 +530,10 @@ class NeoHub:
         NeoStats that are in timeclock mode.
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
 
         message = {"TIMER_ON" if state else "TIMER_OFF": names}
         reply = {"result": "timers on" if state else "timers off"}
@@ -482,7 +548,10 @@ class NeoHub:
         This function works with NeoStats in timeclock mode
         """
 
-        names = [x.name for x in devices]
+        try:
+            names = [x.name for x in devices]
+        except (TypeError, AttributeError):
+            raise NeoHubUsageError('devices must be a list of NeoStat objects')
 
         message = {"TIMER_HOLD_ON" if state else "TIMER_HOLD_OFF": [minutes, names]}
         reply = {"result": "timer hold on" if state else "timer hold off"}
